require "thread"

$LOAD_PATH.push(File.dirname(__FILE__)+"/lib") 
$LOAD_PATH.push(File.dirname(__FILE__)+"/class") 
require 'mainwindow_design.rb'
require 'pattern_plotter.rb'
require 'gerber_reader.rb'
require 'gcode_formatter.rb'
require 'form_ProgressDlg.rb'
require 'form_BuildDlg.rb'
require 'form_PreferenceDlg.rb'
require 'preference.rb'


# --------------------------------------
# --- application class
# --------------------------------------

class AppGerberConverter < GgcMainWindow

  def initialize
    super
    @canvas = PatternPlotter.new
    @canvas.set_round_resolution(12)
    @canvas.marge_outline = false
    @g_pattern = nil
    @g_outline = nil
    @g_drill = nil
    @pattern_build = []
    @pattern_outline = []
    @offset_x = 0.0
    @offset_y = 0.0
    # - preference
    @pref = PreferenceHolder.new
    @pref.load
  end

  # -----------------------------
  # --- menu event
  # -----------------------------
  class BuildProgress
    def initialize
      @jobnum = 0
      @progress = 0
      @passnum = 0
      @pass = 0
      @done = false
      @item_tri = nil
      @item_line = nil
      @item_overline = []
    end

    attr_accessor :jobnum
    attr_accessor :progress
    attr_accessor :passnum
    attr_accessor :pass
    attr_accessor :done
    attr_accessor :item_tri
    attr_accessor :item_line
    attr_accessor :item_overline
  end

  # --- load pattern gerber data
  def menu_open_pattern
    # - preference
    fail = false
    p_resolution = @pref.get('resolution').to_i
    p_milsize = @pref.get('cuttersize').to_f
    fail = true if p_resolution < 4
    fail = true if p_milsize <= 0.0
    if fail then
      System::Windows::Forms::MessageBox.Show("Parameter Error!")
      return
    end
    #
    dlg = System::Windows::Forms::OpenFileDialog.new
    dlg.RestoreDirectory = true
    return if dlg.ShowDialog != System::Windows::Forms::DialogResult.OK
    #
    filename = dlg.FileName
    glscreen.clear_pattern_object
    @canvas.clear_pattern
    @canvas.marge_outline = false
    @canvas.set_round_resolution((p_resolution/2).to_i)
    @g_pattern = GFileReader.new
    if (@g_pattern.load(filename) == true) then
      workmes = Queue.new
      t = Thread.new do
        count = 0
        is_polygon = false
        polyline = []
        polypitch = 0.15
        @g_pattern.get.each do |item|
          if workmes.size == 0 then
            to_progress = BuildProgress.new
            to_progress.jobnum = @g_pattern.get.size
            to_progress.progress = count
            to_progress.passnum = 1
            to_progress.pass = 1
            if (count % 10) == 0 then
              to_progress.item_tri = @canvas.get_object
              to_progress.item_line = @canvas.get_outline
            end
            workmes.push(to_progress)
          end
          count += 1
          #
          if (item['polygon'] == false) && (is_polygon == true) then
            is_polygon = false
            @canvas.set_mil_size(p_milsize)
            @canvas.set_pad_round(true)
            @canvas.set_pad_size(polypitch)
            @canvas.draw_polygon(polyline)
            polyline = []
          elsif (item['polygon'] == true) && (is_polygon == false) then
            is_polygon = true
          end
          #
          if is_polygon == true then
            line = PLine.new
            line.set_point(0, PPoint.new(item['start_x'], item['start_y'], 0))
            line.set_point(1, PPoint.new(item['end_x'], item['end_y'], 0))
            polyline << line
          else
            @canvas.set_mil_size(p_milsize)
            @canvas.set_pad_round(item['is_round'])
            @canvas.set_pad_size(item['width'])
            p1 = PPoint.new(item['start_x'], item['start_y'], 0)
            p2 = PPoint.new(item['end_x'], item['end_y'], 0)
            if (p1.x == p2.x) && (p1.y == p2.y) then
              @canvas.flash(p1.x, p1.y)
            else
              @canvas.move(p1.x, p1.y, p2.x, p2.y)
            end
          end
        end
        # polygon follow
        if (is_polygon == true) && (polyline.size > 0) then
          is_polygon = false
          @canvas.set_mil_size(p_milsize)
          @canvas.set_pad_round(true)
          @canvas.set_pad_size(polypitch)
          @canvas.draw_polygon(polyline)
          polyline = []
        end
        # thread end
        to_progress = BuildProgress.new
        to_progress.jobnum = @g_pattern.get.size
        to_progress.progress = count
        to_progress.done = true
        workmes.push(to_progress)
      end
      progressdlg = FormProgressDialog.new(workmes, glscreen)
      if progressdlg.show_dialog() == System::Windows::Forms::DialogResult.Cancel then
        t.kill
      end

      @canvas.get_object.each do |tri|
        glscreen.add_object(tri)
      end
      glscreen.view_reset

      if @g_pattern.get.size == 0 then
        System::Windows::Forms::MessageBox.Show("Pattern file is empty or wrong format.")
        @g_pattern = nil
      end
    else
      System::Windows::Forms::MessageBox.Show("Pattern file is wrong format.")
      @g_pattern = nil
    end
  end

  # --- load cutting outline
  def menu_open_outline
    dlg = System::Windows::Forms::OpenFileDialog.new
    dlg.RestoreDirectory = true
    return if dlg.ShowDialog != System::Windows::Forms::DialogResult.OK
    filename = dlg.FileName
    glscreen.clear_outline_object
    @g_outline = GFileReader.new
    if (@g_outline.load(filename) == true) then
      gaikei = GCodeFormatter.new
      @g_outline.get.each do |item|
        gaikei.add_line(item['start_x'], item['start_y'],
                        item['end_x'], item['end_y'])
        line = PLine.new
        line.set_point(0, PPoint.new(item['start_x'], item['start_y'], 0))
        line.set_point(1, PPoint.new(item['end_x'], item['end_y'], 0))
        glscreen.add_outline(line)
      end
      gaikei.calc_offset
      @offset_x = gaikei.offset_x
      @offset_y = gaikei.offset_y
      glscreen.view_reset

      if @g_outline.get.size == 0 then
        System::Windows::Forms::MessageBox.Show("Outline file is empty or wrong format.")
        @g_outline = nil
      end
    else
      System::Windows::Forms::MessageBox.Show("Outline file is wrong format.")
      @g_outline = nil
    end
  end

  # --- load drill data
  def menu_open_drill
    dlg = System::Windows::Forms::OpenFileDialog.new
    dlg.RestoreDirectory = true
    return if dlg.ShowDialog != System::Windows::Forms::DialogResult.OK
    @g_drill = dlg.FileName
    glscreen.clear_drill_object
    drl = GCodeDrill.new
    drl.scale = 10 ** @pref.get('drillinputfix').to_i
    drl.dwell = @pref.get('drilldwell').to_i
    drl.load( @g_drill )
    drl.position.each do |hole|
      glscreen.add_point(PPoint.new(hole['x'], hole['y'], 0))
    end
    glscreen.view_reset
  end

  # --- export G-Code
  def menu_export
    # - preference
    fail = false
    p_cuttersize = @pref.get('cuttersize').to_f
    p_patternfeed = @pref.get('patternfeed').to_f
    p_patterndepth = @pref.get('patterndepth').to_f
    p_outlinefeed = @pref.get('outlinefeed').to_f
    p_outlinelayer = @pref.get('outlinelayer').to_f
    p_boardthin = @pref.get('boardthin').to_f
    p_drilldepth = @pref.get('drilldepth').to_f
    fail = true if p_patternfeed <= 0.0
    fail = true if p_patterndepth <= 0.0
    fail = true if p_outlinefeed <= 0.0
    fail = true if p_outlinelayer <= 0.0
    fail = true if p_boardthin <= 0.0
    fail = true if p_drilldepth <= 0.0
    if fail then
      System::Windows::Forms::MessageBox.Show("Parameter Error!")
      return
    end
    if p_outlinelayer > p_boardthin then
      p_outlinelayer = p_boardthin
    end
    # - select dir
    dlg = System::Windows::Forms::FolderBrowserDialog.new
    return if dlg.ShowDialog != System::Windows::Forms::DialogResult.OK
    dir = dlg.SelectedPath
    #
    @offset_x = 0.0
    @offset_y = 0.0
    is_first = true
    # - gaikei
    gaikei = GCodeFormatter.new
    if @g_outline != nil then
      is_first = false
      @g_outline.get.each do |item|
        gaikei.add_line(item['start_x'], item['start_y'],
                        item['end_x'], item['end_y'])
      end
      gaikei.calc_offset
      @offset_x = gaikei.offset_x
      @offset_y = gaikei.offset_y
      gaikei.float_height = 1.0
      gaikei.mil_feed = p_outlinefeed
      gaikei.mil_depth = -p_boardthin
      gaikei.layer_height = -p_outlinelayer
      gaikei.save(dir+"/nc_gaikei.ncd")
    end
    # - pattern
    if @pattern_build.size > 0 then
      gcf = GCodeFormatter.new
      @pattern_build.each do |pat|
        pat.each do |line|
          gcf.add_line(line[0].x, line[0].y, line[1].x, line[1].y)
        end
      end
      gcf.calc_offset
      if (gcf.offset_x < @offset_x) || (gcf.offset_y < @offset_y) then
        if (gcf.offset_x < @offset_x) then
          @offset_x = gcf.offset_x
        end
        if (gcf.offset_y < @offset_y) then
          @offset_y = gcf.offset_y
        end
        if @g_outline != nil then
          gaikei.offset_x = @offset_x
          gaikei.offset_y = @offset_y
          gaikei.rebuild
          gaikei.save(dir+"/nc_gaikei.ncd")
        end
      end
      gcf.offset_x = @offset_x
      gcf.offset_y = @offset_y
      gcf.mil_feed = p_patternfeed
      gcf.float_height = 0.5
      gcf.mil_depth = -p_patterndepth
      if is_first == true then
        gcf.calc_offset
        @offset_x = gcf.offset_x
        @offset_y = gcf.offset_y
        is_first = false
      end
      gcf.near_collect = (p_cuttersize / 4.0)
      gcf.save(dir+"/nc_pattern.ncd")
    end
    # - drill
    if @g_drill != nil then
      drl = GCodeDrill.new
      drl.scale = 10 ** @pref.get('drillinputfix').to_i
      drl.dwell = @pref.get('drilldwell').to_i
      drl.offset_x = @offset_x
      drl.offset_y = @offset_y
      drl.mil_depth = -p_drilldepth
      drl.float_height = 0.5
      drl.load( @g_drill )
      drl.save(dir+"/nc_drl.ncd")
    end
  end

  # --- set preference
  def menu_preference
    dlg = FormPreferenceDialog.new
    dlg.num_resolution   = @pref.get('resolution').to_i
    dlg.num_pass         = @pref.get('pass').to_i
    dlg.num_cuttersize   = @pref.get('cuttersize').to_f
    dlg.num_patternfeed  = @pref.get('patternfeed').to_f
    dlg.num_patterndepth = @pref.get('patterndepth').to_f
    dlg.num_outlinefeed  = @pref.get('outlinefeed').to_f
    dlg.num_outlinelayer = @pref.get('outlinelayer').to_f
    dlg.num_boardthin    = @pref.get('boardthin').to_f
    dlg.num_drilldepth   = @pref.get('drilldepth').to_f
    dlg.num_drilldwell   = @pref.get('drilldwell').to_i
    dlg.num_drillinputfix = @pref.get('drillinputfix').to_i
    dlg.num_threadnum = @pref.get('threadnum').to_i
    if dlg.ShowDialog == System::Windows::Forms::DialogResult.OK then
      @pref.set('resolution',   dlg.num_resolution.to_s)
      @pref.set('pass',         dlg.num_pass.to_s)
      @pref.set('cuttersize',   dlg.num_cuttersize.to_s)
      @pref.set('patternfeed',  dlg.num_patternfeed.to_s)
      @pref.set('patterndepth', dlg.num_patterndepth.to_s)
      @pref.set('outlinefeed',  dlg.num_outlinefeed.to_s)
      @pref.set('outlinelayer', dlg.num_outlinelayer.to_s)
      @pref.set('boardthin',    dlg.num_boardthin.to_s)
      @pref.set('drilldepth',   dlg.num_drilldepth.to_s)
      @pref.set('drilldwell',   dlg.num_drilldwell.to_s)
      @pref.set('drillinputfix', dlg.num_drillinputfix.to_s)
      @pref.set('threadnum', dlg.num_threadnum.to_s)
    end
  end

  # --- build carve path
  def menu_build
    # - preference
    fail = false
    p_resolution = @pref.get('resolution').to_i
    p_overpath = @pref.get('pass').to_i
    p_milsize = @pref.get('cuttersize').to_f
    p_threadnum = @pref.get('threadnum').to_i
    fail = true if p_resolution < 4
    fail = true if p_overpath < 1
    fail = true if p_milsize <= 0.0
    fail = true if p_threadnum < 1
    if fail then
      System::Windows::Forms::MessageBox.Show("Parameter Error!")
      return
    end
    # - confirm
    dlg = FormBuildDialog.new(p_overpath, p_threadnum)
    if dlg.do == System::Windows::Forms::DialogResult.Cancel then
      return
    end
    start_time = Time::now()
    p_overpath = dlg.value_pass.to_i
    p_threadnum = dlg.value_thread.to_i
    #
    @canvas.clear_pattern
    @canvas.marge_outline = true
    @canvas.set_round_resolution((p_resolution/2).to_i)
    @pattern_build = []
    workmes = Queue.new
    t = Thread.new do
      for overmil in 1..p_overpath do
        count = 0
        if overmil == 1 then
          milsize = p_milsize
        else
          milsize = p_milsize + (p_milsize * 0.6 * (overmil-1))
        end
        # - add pattern
        is_polygon = false
        polyline = []
        polypitch = 0.15
        @g_pattern.get.each do |item|
          if workmes.size == 0 then
            to_progress = BuildProgress.new
            to_progress.jobnum = @g_pattern.get.size * 9
            to_progress.progress = count
            to_progress.passnum = p_overpath
            to_progress.pass = overmil
            to_progress.item_tri = @canvas.get_object
            #to_progress.item_line = @canvas.get_outline
            to_progress.item_overline = @pattern_build
            workmes.push(to_progress)
          end
          count += 1
          #
          if (item['polygon'] == false) && (is_polygon == true) then
            is_polygon = false
            @canvas.set_mil_size(p_milsize)
            @canvas.set_pad_round(true)
            @canvas.set_pad_size(polypitch)
            @canvas.draw_polygon(polyline)
            polyline = []
          elsif (item['polygon'] == true) && (is_polygon == false) then
            is_polygon = true
          end
          #
          if is_polygon == true then
            line = PLine.new
            line.set_point(0, PPoint.new(item['start_x'], item['start_y'], 0))
            line.set_point(1, PPoint.new(item['end_x'], item['end_y'], 0))
            polypitch = item['width']
            polyline << line
          else
            @canvas.set_mil_size(p_milsize)
            @canvas.set_pad_round(item['is_round'])
            @canvas.set_pad_size(item['width'])
            p1 = PPoint.new(item['start_x'], item['start_y'], 0)
            p2 = PPoint.new(item['end_x'], item['end_y'], 0)
            if (p1.x == p2.x) && (p1.y == p2.y) then
              @canvas.flash(p1.x, p1.y)
            else
              @canvas.move(p1.x, p1.y, p2.x, p2.y)
            end
          end
        end
        # polygon follow
        if (is_polygon == true) && (polyline.size > 0) then
          is_polygon = false
          @canvas.set_mil_size(p_milsize)
          @canvas.set_pad_round(true)
          @canvas.set_pad_size(polypitch)
          @canvas.draw_polygon(polyline)
          polyline = []
        end
        #
        to_progress = BuildProgress.new
        to_progress.jobnum = 100
        to_progress.progress = 9
        to_progress.passnum = p_overpath
        to_progress.pass = overmil
        to_progress.item_tri = @canvas.get_object
        to_progress.item_overline = @pattern_build
        workmes.push(to_progress)
        #
        @canvas.prepare_outline(p_threadnum)
        to_progress = BuildProgress.new
        to_progress.jobnum = 100
        to_progress.progress = 10
        to_progress.passnum = p_overpath
        to_progress.pass = overmil
        workmes.push(to_progress)
        # - extruct outline
        one_thread_worknum = 25
        primitive_num = @canvas.get_pattern_num
        geta = primitive_num * 10 / 100
        progress_unit = one_thread_worknum * p_threadnum
        progress_turn = (primitive_num / progress_unit) + 1
        for counter_loop in 0...progress_turn
          itemnum = counter_loop * progress_unit
          tlist = []
          for nthread in 0...p_threadnum
            endnum = itemnum + one_thread_worknum
            if endnum > primitive_num then
              endnum = primitive_num
            end
            if endnum > itemnum then 
              t = Thread.new(itemnum, endnum) do |startnum, tonum|
                for item_loop in startnum...tonum
                  @canvas.extruct_outline(item_loop)
                end
              end
              tlist << t
            end
            itemnum += one_thread_worknum
            #break if itemnum >= primitive_num 
          end
          tlist.each do |t|
            t.join
          end
          #
          donenum = (counter_loop + 1) * progress_unit
          if donenum > primitive_num then
            donenum = primitive_num
          end
          if (workmes.size == 0) then
            to_progress = BuildProgress.new
            to_progress.jobnum = primitive_num + geta
            to_progress.progress = donenum + geta
            to_progress.passnum = p_overpath
            to_progress.pass = overmil
            to_progress.item_tri = @canvas.get_object
            to_progress.item_line = @canvas.get_outline_bound(donenum)
            to_progress.item_overline = @pattern_build
            workmes.push(to_progress)
          end
        end
        # result
        if @canvas.get_outline.size > 0 then
          @pattern_build << @canvas.get_outline
        end
        if overmil != p_overpath then
          @canvas.clear_pattern
        end
      end
      # thread end
      to_progress = BuildProgress.new
      to_progress.jobnum = @g_pattern.get.size
      to_progress.progress = count
      to_progress.done = true
      workmes.push(to_progress)
    end
    progressdlg = FormProgressDialog.new(workmes, glscreen)
    if progressdlg.show_dialog() == System::Windows::Forms::DialogResult.Cancel then
      t.kill
    end
    
    glscreen.clear_pattern_object
    @canvas.get_object.each do |tri|
      glscreen.add_object(tri)
    end
    @pattern_build.each do |pat|
      pat.each do |line|
        glscreen.add_line(line)
      end
    end
    glscreen.view_reset
    #
    working_time = Time::now() - start_time;
    System::Windows::Forms::MessageBox.Show("Gynostemma done.\nElapsing time "+Time.at(working_time).getgm.strftime("%H:%M:%S"))    
  end

  # --- eraser
  def menu_clean
    @g_pattern = nil
    @g_outline = nil
    @g_drill = nil
    @pattern_build.clear
    @pattern_outline.clear
    @canvas.clear_pattern
    glscreen.clear_pattern_object
    glscreen.clear_outline_object
    glscreen.clear_drill_object
    glscreen.view_reset
  end

  # ---------------------------
  # --- closed work
  def on_close
    @pref.save
    super
  end

end




# --------------------------------------
# --- execute
# --------------------------------------

app = AppGerberConverter.new
app.run

